from django.core import checks
from django.core.exceptions import ValidationError
from django.db import models
from django.forms import ModelForm
from django.test import TestCase
from django.utils.datastructures import MultiValueDict

from attachments.models import Attachment

from ..forms import BBCodeModelForm
from ..models import BBCode, BBCodeField
from .models import TestModel, TestModelNoAttach, TestModelAttachNotNullable


class IncorrectModelForm(ModelForm):
    class Meta:
        model = TestModel
        fields = ["bbcode"]


class IncorrectModelForm2(BBCodeModelForm):
    class Meta:
        model = TestModel
        fields = ["bbcode"]


class IncorrectModelForm3(BBCodeModelForm):
    class Meta:
        model = TestModel
        fields = ["bbcode"]
        bbcode_field = "aa"


class TestModelForm(BBCodeModelForm):
    class Meta:
        model = TestModel
        fields = ["bbcode"]
        bbcode_field = "bbcode"


class TestModelFormWithAttachTarget(BBCodeModelForm):
    class Meta:
        model = TestModel
        fields = ["bbcode"]
        bbcode_field = "bbcode"
        main_attach_target = "main_attachment"


class TestModelFormInvalidTarget(BBCodeModelForm):
    class Meta:
        model = TestModel
        fields = ["bbcode"]
        bbcode_field = "bbcode"
        main_attach_target = "invalid_target"


class TestModelFormAttachTargetNotNullable(BBCodeModelForm):
    class Meta:
        model = TestModelAttachNotNullable
        fields = ["bbcode", "main_attachment"]
        bbcode_field = "bbcode"
        main_attach_target = "main_attachment"


class TestModelFormNoAttach(BBCodeModelForm):
    class Meta:
        model = TestModelNoAttach
        fields = ["bbcode"]
        bbcode_field = "bbcode"


class TestModelFormNoAttachWithTarget(BBCodeModelForm):
    class Meta:
        model = TestModelNoAttach
        fields = ["bbcode"]
        bbcode_field = "bbcode"
        main_attach_target = "some_target"


class FakeAttachmentManager(object):
    def __init__(self, test):
        self.uploaded_attachments = []
        self._orig_manager = None
        self._test = test

    def __enter__(self):
        self._orig_manager = Attachment.objects
        Attachment.objects = self

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        Attachment.objects = self._orig_manager

    def upload_file(self, namespace, attachment):
        self._test.assertEqual(namespace, "test_namespace")

        if attachment == "file1":
            self.uploaded_attachments.append("file1")
        elif attachment == "file2":
            raise ValidationError(["Validation", "Error!"])
        else:  # pragma: no cover
            self._test.fail()

    def get_available_attachments_for(self, namespace, owner_id):
        self._test.assertEqual(namespace, "test_namespace")

        return [
            Attachment(
                file_name="some file.txt",
                ident_main="test1",
                ident_mini=None,
                file_type="txt",
                file_size=42,
                namespace="test_namespace",
            )
        ]


class BBCodeDatabaseFieldsTestCase(TestCase):
    def test_check_invalid_pack(self):
        # On déclare la classe dans le test pour que les checks faits par
        # Django au début des tests n'échouent pas

        class TestModelInvalidPack(models.Model):
            bbcode = BBCodeField(pack="invalid")

        errors = TestModelInvalidPack.check()
        self.assertEqual(
            errors,
            [
                checks.Error(
                    "Invalid pack attribute invalid. Must be one of: message, "
                    "message_attach, web_page, web_page_attach",
                    hint=None,
                    obj=TestModelInvalidPack._meta.fields[1],
                    id="bbcode.E001",
                ),
            ],
        )

    def test_save_none(self):
        object = TestModelNoAttach()
        object.bbcode = None
        object.save()

        self.assertEqual(object.id, 1)
        self.assertIs(object.bbcode, None)

    def test_save_string(self):
        object = TestModelNoAttach()
        object.bbcode = "[b]Test[/b]"
        object.save()

        object.refresh_from_db()

        self.assertEqual(object.id, 1)
        self.assertIsInstance(object.bbcode, BBCode)

    def test_set_object(self):
        object = TestModelNoAttach()
        with self.assertRaises(ValueError):
            object.bbcode = ()
            object.save()

    def test_attempt_lookup(self):
        with self.assertRaises(TestModelNoAttach.DoesNotExist):
            TestModelNoAttach.objects.get(bbcode="abc")
        obj = TestModelNoAttach(bbcode="abc")
        obj.save()
        looked_up = TestModelNoAttach.objects.get(bbcode="abc")
        self.assertEqual(obj, looked_up)


class BBCodeFormsTestCase(TestCase):
    def test_incorrect_model_form_not_bbcode(self):
        form = IncorrectModelForm(dict(bbcode="abc"))
        self.assertTrue(form.is_valid())
        with self.assertRaisesRegex(RuntimeError, r"^Ce BBCode utilise des "):
            form.save()

    def test_incorrect_model_form_no_fields_parameter(self):
        with self.assertRaisesRegex(ValueError, r"has no bbcode_field meta"):
            IncorrectModelForm2(dict(bbcode="abc"), namespace="test_namespace")

    def test_incorrect_model_form_wrong_fields_parameter(self):
        with self.assertRaisesRegex(ValueError, r"BBCode field 'aa' is not "):
            IncorrectModelForm3(dict(bbcode="abc"), namespace="test_namespace")

    def test_model_form_no_namespace(self):
        with self.assertRaisesRegex(TypeError, r"^Missing argument 'namespace'$"):
            TestModelForm(dict(bbcode="abc"))

    def test_invalid_bbcode(self):
        form = TestModelForm(dict(bbcode="[b]abc"), namespace="test_namespace")
        self._check_form_error(form, "BBCode [b] mal fermé.")

    def test_nonexistent_attachment(self):
        form = TestModelForm(dict(bbcode="[attach]abc[/attach]"), namespace="test_namespace")
        self._check_form_error(form, "La pièce jointe abc n’est pas " "disponible.")

        instance = TestModel()
        instance.id = 1
        form = TestModelForm(
            dict(bbcode="[attach]abc[/attach]"),
            namespace="test_namespace",
            instance=instance,
        )
        self._check_form_error(form, "La pièce jointe abc n’est pas " "disponible.")

    def test_existing_attachment(self):
        mgr = Attachment.objects
        items = mgr.get_available_attachments_for("test_namespace")
        self.assertEqual(len(items), 0)

        attachment = mgr.create(
            file_name="some file.txt",
            ident_main="abcdefgh",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        items = mgr.get_available_attachments_for("test_namespace")
        self.assertEqual(len(items), 1)

        form = TestModelForm(dict(bbcode="[attach]abcdefgh[/attach]"), namespace="test_namespace")

        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(Attachment.objects.get(pk=attachment.pk).owner_id, 1)

    def test_multiple_attachments_with_target(self):
        mgr = Attachment.objects

        mgr.create(
            file_name="attach1.txt",
            ident_main="attach1",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        attachment2 = mgr.create(
            file_name="attach2.png",
            ident_main="attach2",
            ident_mini="attach2mini",
            file_type="png",
            file_size=42,
            namespace="test_namespace",
        )

        mgr.create(
            file_name="attach3.txt",
            ident_main="attach3",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        mgr.create(
            file_name="attach4.png",
            ident_main="attach4",
            ident_mini="attach4mini",
            file_type="png",
            file_size=42,
            namespace="test_namespace",
        )

        message = """
            [attach]attach1[/attach]
            [attach]attach2[/attach]
            [attach]attach3[/attach]
            [attach]attach4[/attach]
        """

        form = TestModelFormWithAttachTarget(dict(bbcode=message), namespace="test_namespace")
        instance = form.save()

        self.assertEqual(instance.main_attachment, attachment2)

    def test_multiple_attachments_with_target_no_commit(self):
        mgr = Attachment.objects

        mgr.create(
            file_name="attach1.txt",
            ident_main="attach1",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        attachment2 = mgr.create(
            file_name="attach2.png",
            ident_main="attach2",
            ident_mini="attach2mini",
            file_type="png",
            file_size=42,
            namespace="test_namespace",
        )

        mgr.create(
            file_name="attach3.txt",
            ident_main="attach3",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        mgr.create(
            file_name="attach4.png",
            ident_main="attach4",
            ident_mini="attach4mini",
            file_type="png",
            file_size=42,
            namespace="test_namespace",
        )

        message = """
            [attach]attach1[/attach]
            [attach]attach2[/attach]
            [attach]attach3[/attach]
            [attach]attach4[/attach]
        """

        form = TestModelFormWithAttachTarget(dict(bbcode=message), namespace="test_namespace")
        instance = form.save(commit=False)

        self.assertEqual(instance.main_attachment, attachment2)

    def test_multiple_attachments_with_target_no_image(self):
        mgr = Attachment.objects

        mgr.create(
            file_name="attach1.txt",
            ident_main="attach1",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        mgr.create(
            file_name="attach2.txt",
            ident_main="attach2",
            ident_mini=None,
            file_type="txt",
            file_size=42,
            namespace="test_namespace",
        )

        message = """
            [attach]attach1[/attach]
            [attach]attach2[/attach]
        """

        form = TestModelFormWithAttachTarget(dict(bbcode=message), namespace="test_namespace")
        instance = form.save()

        self.assertEqual(instance.main_attachment, None)

    def test_no_attachments(self):
        with self.assertRaisesRegex(Exception, r"is not needed"):
            TestModelFormNoAttach(dict(bbcode="[b]abcdefgh[/b]"), namespace="test_namespace")

        with self.assertRaisesRegex(Exception, r"is not needed"):
            TestModelFormNoAttachWithTarget(dict(bbcode="[b]abcdefgh[/b]"))

        with self.assertRaisesRegex(ValueError, r"is not an FK"):
            TestModelFormInvalidTarget(dict(bbcode="[b]abcdefgh[/b]"), namespace="test_namespace")

        with self.assertRaisesRegex(ValueError, r"is not nullable"):
            TestModelFormAttachTargetNotNullable(dict(bbcode="[b]abcdefgh[/b]"), namespace="test_namespace")

        form = TestModelFormNoAttach(dict(bbcode="[b]abcdefgh[/b]"))

        self.assertTrue(form.is_valid())

        with self.assertRaisesRegex(Exception, r"has no attachments"):
            form.attachments_fieldset()

        with self.assertRaisesRegex(Exception, r"are not supported with"):
            form.cleaned_data["bbcode"].validate_attachments(None, None)

        with self.assertRaisesRegex(Exception, r"are not supported with"):
            form.cleaned_data["bbcode"].commit_attachments(None, None)

        with self.assertRaisesRegex(Exception, r"are not supported with"):
            form.cleaned_data["bbcode"].get_main_attachment()

        form.save()

    def test_render(self):
        form = TestModelForm(dict(bbcode=None), namespace="test_namespace")
        assert '<textarea name="bbcode" rows="10" cols="40" id="id_bbcode"' in form.as_p()

    def test_upload_no_files(self):
        with FakeAttachmentManager(self):
            data = {
                "bbcode": "[b]",
                "submit_attachments": "<Something…>",
            }
            uploads = MultiValueDict()
            form = TestModelForm(data, uploads, namespace="test_namespace")
            self.assertFalse(form.is_valid())
            self.assertEqual(form.errors, {})

            self.assertIsNone(form.get_preview())

    def test_upload_multiple_files_with_error(self):
        with FakeAttachmentManager(self) as mgr:
            data = {
                "bbcode": "[b]",
                "submit_attachments": "<Something…>",
            }
            uploads = MultiValueDict(
                {
                    "attachment_upload[]": [
                        "file1",
                        "file2",
                    ]
                }
            )
            form = TestModelForm(data, uploads, namespace="test_namespace")
            self.assertFalse(form.is_valid())
            self.assertEqual(form.errors, {})

            self.assertIsNone(form.get_preview())

            self.assertEqual(mgr.uploaded_attachments, ["file1"])
            self.assertEqual(form.attachments_errors, ["Validation", "Error!"])

    def test_attach_fieldset(self):
        with FakeAttachmentManager(self) as mgr:
            data = {
                "bbcode": "[b]",
            }
            uploads = MultiValueDict()
            form = TestModelForm(data, uploads, namespace="test_namespace")
            self.assertFalse(form.is_valid())
            self.assertEqual(list(form.errors.keys()), ["bbcode"])

            self.assertIsNone(form.get_preview())

            self.assertEqual(mgr.uploaded_attachments, [])
            self.assertEqual(form.attachments_errors, [])

            self.assertInHTML(
                """
                <div class="attachment"
                onclick="insertAttachment('id_bbcode', 'test1')">
                <img src="/static/bbcode/file_extension_txt.png" />
                <div>some file.txt</div>
                Taille : 42 octets
                </div>
            """,
                form.attachments_fieldset(),
            )

    def test_preview(self):
        data = {
            "bbcode": "[b]Hello, world![/b]",
        }
        form = TestModelForm(data, namespace="test_namespace")
        self.assertTrue(form.is_valid())
        self.assertFalse(form.errors)

        self.assertHTMLEqual(
            form.get_preview(),
            """
            <b>Hello, world!</b>
        """,
        )

    def _check_form_error(self, form, err_msg):
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, dict(bbcode=[err_msg]))
